/*
 * s02q2.java
 *
 * Created on February 13, 2002, 3:22 PM
 */

/**
 *
 * @author  Chris Schmidt
 */

import java.io.*;
import java.util.*;

import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;

import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;

import java.util.regex.*;

/** Primary class that parses an XML file and prints to stdout an HTML detailing information on the XML file.
 */
public class s02q2 extends DefaultHandler {
    // Class variables
    private XMLPath xmlPath;
    private ElementTable elementTable;
    private AttributeTable attributeTable;
    private Locator loc;
    private int elementId;
    private StringBuffer fileContents;
    private int PIlength;
    private int iEndElement;
    
    /** Creates a new instance of s02q2 */
    public s02q2() {
        xmlPath = new XMLPath();
        elementTable = new ElementTable();
        attributeTable = new AttributeTable();
        elementId = 0;
        fileContents = new StringBuffer();
        PIlength = 0;
        iEndElement = 0;
    }
    
    /** Main function, start program here.
     * @param argv Command line arguments sent into the main function.
     */    
    public static void main(String argv[]) {
        // Local Variables
        File xmlFile;
        String sFilename;
        // Use an instance of ourselves as the SAX event handler
        s02q2 handler = new s02q2();
        
        // Begin handler and read in the file sent in by the user
        // Check that user passed in a filename (or something)
        if (argv.length != 1) {
            System.err.println("Error: Improper number of arguments.\n" +
                               "Usage: java s02q2 filename\nWhere filename is a" +
                               " XML file you wish to parse.");
            System.exit(-1);
        }
        
        // Get the file that the user wishes to parse and manipulate it
        sFilename = argv[0];
        xmlFile = handler.manipulateXMLFile(new File(sFilename));                

        // Use the default (non-validating) parser
        SAXParserFactory factory = SAXParserFactory.newInstance();
        try {
            // Parse the input
            SAXParser saxParser = factory.newSAXParser();
            saxParser.parse( xmlFile, (DefaultHandler)handler );
        } catch (Throwable t) {
            t.printStackTrace();
        }
        
        // At this point, the information should have been collected, output
        // the data.
        handler.printData();

        
    }
    
    /** Called by the SAX parser when an element is discovered while going through the XML file
     * @param uri
     * @param localName
     * @param qName
     * @param attributes
     * @throws SAXException
     */    
    public void startElement(String uri,String localName,String qName,Attributes attributes) throws SAXException {
        int colNum = loc.getColumnNumber();
        int numAttributes;        
        elementId++;
        
        xmlPath.pushElement(qName);
        xmlPath.addElement(qName,attributes);
        elementTable.addElement(qName,xmlPath.getPathIndex(null),iEndElement);
        // Set the iEndElement variable so the next element can use the current
        // column number as its start number
        iEndElement = colNum;
        // Determine if there are attributes and add its information
        numAttributes = attributes.getLength();
        
        for (int loop = 0; loop < numAttributes; loop++) {
            attributeTable.addAttributes(attributes.getQName(loop),
              attributes.getValue(loop),elementId,qName,
              xmlPath.getPathIndex(attributes.getQName(loop)));
        }       
    }
    
    /** Called when the SAX parser encounters an end tag for an element.
     * @param namespaceURI
     * @param localName
     * @param qName
     */    
    public void endElement(String namespaceURI, String localName, String qName) {
        // Finish the element table information before removing the element from
        // the path's stack
        int colNum = loc.getColumnNumber();
        
        elementTable.finishElement(qName,xmlPath.getPathIndex(null),colNum);
        // Set the iEndElement variable so the next element can use the current
        // column number as its start number
        iEndElement = colNum;
        
        xmlPath.popElement(qName);
    }
    
    /** Called by the SAX parser at the beginning of parsing to set a reference for the document locator variable.
     * @param locator  */    
    public void setDocumentLocator(Locator locator) {
        loc = locator;
    }
  
    /**
     *  Method responsible for printing the HTML data to standard out.  This method 
     *  is called after all parsing is done and information collected.
     */
    private void printData() {
        // Prints the data collected during the parse of the XML file
        int count;
        String path;
        ElementTableData elementData;
        AttributeTableData attrData;
        
        // Print the header information (including the creation stats)
        System.out.println("<html><head><title>XML tables</title></head>\n");
        System.out.println("<body style=\"font-family:verdana;background-color:#ccccff\">\n");
        System.out.println("<h3>Created Time</h3>"+new Date()+"\n");
        // Print the source of the XML file
        System.out.println("<h3>Source XML</h3>\n");
        System.out.print("<div style=\"border:1px solid black; padding:5px\">");
        System.out.println(fileContents + "</div>");
        // Print the path table
        System.out.println("<h3>Path Table</h3>\n");
        System.out.println("<table border=\"1\"><tr><td>Id</td><td>Path Expressions</td></tr>\n");        
        // -- Break out the path list into a format for the table
        count = 0;
        for (Enumeration e = xmlPath.getPathList(); e.hasMoreElements();) {
            count++;
            path = (String)e.nextElement();
            System.out.println("<tr><td>" + count + "</td><td>" + path + "</td></tr>\n");            
        }
        System.out.println("</table>\n");
        // Print the element table
        System.out.println("<h3>Element Table</h3>\n");
        System.out.println("<table border=\"1\"><tr><td>Id</td><td>Name</td><td>Path Id</td><td>start index</td><td>end index</td></tr>\n");
        // -- Break out the element table into the proper format for the HTML table
        count = 0;
        for (Enumeration e = elementTable.getElements(); e.hasMoreElements();) {
            count++;
            elementData = (ElementTableData)e.nextElement();
            System.out.println("<tr><td>" + count + "</td><td>" + 
                elementData.getElementName() + "</td><td>" + elementData.getElementIndex() +
                "</td><td>" + elementData.getElementStart() + "</td><td>" +
                elementData.getElementEnd() + "</td></tr>\n");        
        }        
        System.out.println("</table>\n");
        System.out.println("<h3>Attribute Table</h3>\n");
        System.out.println("<table border=\"1\"><tr><td>Name</td><td>Value</td><td>Element Id</td><td>Element Name</td><td>Path Id</td></tr>\n");       
        // -- Break out the attribute table into the proper format for the HTML table
        count = 0;
        for (Enumeration e = attributeTable.getAttributes(); e.hasMoreElements();) {
            count++;
            attrData = (AttributeTableData)e.nextElement();
            System.out.println("<tr><td>"+ attrData.getAttrName() + "</td><td>" +
                attrData.getAttrValue() + "</td><td>" + attrData.getElementId() +
                "</td><td>" + attrData.getElementName() + "</td><td>" + 
                attrData.getPathId() + "</td></tr>\n");        
        }       
        System.out.println("</table>\n");
        System.out.println("</body></html>\n");
    }
    
    /**
     * This method is responsible for loading the target XML file and manipulate
     * it so that it no longer has extraneous whitespaces.  It also sets the 
     * final XML source into memory for later use by the method that prints 
     * all XML information.
     */
    private File manipulateXMLFile(File origFile) {
        // Local variables        
        BufferedReader bufIn = null;
        BufferedWriter bufOut = null;
        String scratchFile = ".scratch.xml";
        File outFile;
        StringBuffer tempStrBuf;
        
        // Read the file into a string buffer for manipulation
        try {
            bufIn = new BufferedReader(new FileReader(origFile));
        } catch (java.io.FileNotFoundException notFound) {
            System.err.println("Error: could not locate file " + 
                    origFile.getPath());
            System.exit(-1);
        }
        
        try {
            while (bufIn.ready()) {
                // This will append the next line to the string buffer,
                // effectively removing all carriage returns
                fileContents.append(bufIn.readLine());
            }
        } catch (java.io.IOException ioErr) {
            System.err.print("Error: ");
            ioErr.printStackTrace(System.err);
        }
        
        // Remove the whitespace between the elements
        // Note: This will remove whitespace between any element
        //  that doesn't include actual character data. I.e. if an
        //  element actually is MEANT to have spaces then this will destroy
        //  that data... which isn't really what we want
        //  This code is based off the code from the javasoft site
        
        tempStrBuf = fileContents;
        // Remove the whitespace at the front and the back
        tempStrBuf = this.simpleReplace(tempStrBuf.toString(),"^\\s*?","");
        tempStrBuf = this.simpleReplace(tempStrBuf.toString(),"\\s*$","");        
        // Variables for removing whitespace
        boolean foundMatch = true;
        boolean result;
        Pattern p = Pattern.compile("(<.*?>)\\s+?(<.*?>)");
        // Create the matcher
        Matcher m;
        
        // Continue to iterate through the buffer trying to match the pattern
        // -- only stop when no matches are found
        // -- this is needed because the pattern sometimes skips over elements
        //    since they are matched in the previous find() call        
        while (foundMatch) {
            StringBuffer sb = new StringBuffer();
            foundMatch = false; 
            m = p.matcher(tempStrBuf);
            
            result = m.find();
            while(result) {
                // Ensure that both groups were found
                // (I have found that it is possible that the match
                //  is found, but the group is not... this seems to be
                //  a defect within the regular expression code.  This
                //  only happens when the XML file is VERY large)
                if (m.groupCount() >= 2) {                    
                    m.appendReplacement(sb, m.group(1) + m.group(2));
                
                    result = m.find();
                    // Found at least one match, need to iterate through the 
                    // top-level while loop at least one more time                    
                    foundMatch = true;
                }
                else {
                    result = false;
                }                
            }
            // Add the last segment of input to 
            // the new String
            m.appendTail(sb);
            // Reset the match to the beginning to check through it again
            tempStrBuf = sb;
        }                        

        // Set file contents with the new whitespace free xml source
        fileContents = tempStrBuf;
        
        // Get the length of the PIs at the start of the XML file
        // Note that there can be more than one PI, so I want the match
        // to be greedy and match over ALL PIs at the start of the file.
        PIlength = getPatternLength(tempStrBuf.toString(),"^<[?].*[?]>");
                
        // The xml doesn't have insignificant whitespace now, push the data into
        // a scratch workfile that the parser can use
        outFile = new File(scratchFile);
        try {
            bufOut = new BufferedWriter(new FileWriter(outFile));
            bufOut.write(fileContents.toString());
            bufOut.flush();
        } catch (IOException ioErrOut) {
            System.err.print("Error: ");
            ioErrOut.printStackTrace(System.err);
        }
        
        // Now update the string buffer to change the '<', '&', etc chars
        // to entities for printing out the xml source later
        // -- Based on the javasoft tutorial on regular expressions:
 
        tempStrBuf = this.simpleReplace(tempStrBuf.toString(),"&","&amp;");
        tempStrBuf = this.simpleReplace(tempStrBuf.toString(),"<","&lt;");
        tempStrBuf = this.simpleReplace(tempStrBuf.toString(),">","&gt;");
        fileContents = tempStrBuf;
        // Return the File that points to the scratch xml file
        return outFile;
    }
    
    /**
     * Simple method to get the length of a resulting pattern from a string
     * This doesn't have too much use in the final program, because I
     * modified the code based on new understanding of the requirements.
     * Originally, this method was used to determing the total length of
     * PIs at the start of the XML document, so I could subtract that 
     * number from the column indices later in the application cycle.
     * That was changed later during development.
     *
     */
    private int getPatternLength(String origStr, String pattern) {
        // Returns the length of a specific pattern to the caller
        // -- This method does not figure out groups in addition
        //    to the main group (i.e. group 0)
        // This is based on the method replace that in turn is based 
        // on code from the javasoft site/reg expr tutorial
        // Create the pattern
        int result = 0;
        Pattern p = Pattern.compile(pattern);
        // Create the matcher
        Matcher m = p.matcher(origStr);
        if (m.find()) {
            // Found a match, return the result
            result = (m.group(0)).length();
        }                
        return result;
    }
    
    /**
     *  Simple method to replace a certain string into another string based on
     *  a regular expression.  This method does not deal with groups, so it
     *  isn't very 'smart' but is useful when you know exactly what you want
     *  to replace with.
     */
    private StringBuffer simpleReplace(String origStr, String pattern,
            String patternRepl) {
        // Takes the original string and replaces all matches of pattern
        // with patternRepl
        // This is based on code from the javasoft site/reg expr tutorial
        // Create a pattern to match pattern
        Pattern p = Pattern.compile(pattern);
        // Create a matcher with an input string
        Matcher m = p.matcher(origStr);
        StringBuffer sb = new StringBuffer();
        boolean result = m.find();
        // Loop through and create a new String 
        // with the replacements
        while(result) {
            m.appendReplacement(sb, patternRepl);
            result = m.find();
        }
        // Add the last segment of input to 
        // the new String
        m.appendTail(sb);
        
        return sb;  
    }
}
